from rest_framework.response import Response
from rest_framework.generics import (CreateAPIView,
                                     GenericAPIView,
                                     ListAPIView)
from rest_framework.serializers import ValidationError
from django_filters.rest_framework import DjangoFilterBackend
from . import serializers
from .models import UniqueVote, ElectionVote
from .filters import UniqueVoteFilter
from .utils import (initialize_election,
                    select_random_entry,
                    save_atomic_entry,
                    close_election,
                    StandardPagination
                    )


class AnonymizeVoteView(CreateAPIView):
    """
    Standard API view
    """
    permission_classes = []
    serializer_class = serializers.AnonymizeVoteSerializer

    def get_serializer_class(self):
        """
        Define the Serializer class to be interpreted
        by Django Swagger
        """
        return self.serializer_class

    def post(self, request, *args, **kwargs):
        """
        POST method docs
        """
        election_id = self.kwargs.get('election_id') or None
        if election_id is None:
            raise ValidationError("No election ID found")
        entry = ElectionVote.objects.filter(election_id=election_id)[:1]

        if not entry.exists():
            raise ValidationError("Election not initialized")

        serializer = serializers.AnonymizeVoteSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        sanitized_data = serializer.validated_data
        try:
            # Create the ElectionVote Object
            entry = ElectionVote()
            entry.ciphered_vote = sanitized_data['ciphered_vote']
            entry.election_id = election_id
            entry.signed_vote_hash = sanitized_data['signed_vote']
            entry.blind_components = sanitized_data['blind_components']
            entry.pr_components = sanitized_data['pr_components']
            entry.empty_entry = False

            # Find a random slot to insert into the DB
            (slot_id, max_nr) = select_random_entry(ElectionVote,
                                                    election_id)

            # Repeat until a valid random slot is designated
            # and the save procedure is successful
            while not save_atomic_entry(entry,
                                        slot_id,
                                        election_id,
                                        max_nr):

                (slot_id, max_nr) = select_random_entry(ElectionVote,
                                                        election_id)

            # Creat the UniqueVote Object
            entry = UniqueVote()
            entry.unique_vote_bulletin = sanitized_data['unique_bulletin']
            entry.election_id = election_id
            entry.empty_entry = False

            # Find a random slot to insert into the DB
            (slot_id, max_nr) = select_random_entry(UniqueVote, election_id)

            # Repeat until a valid random slot is designated
            # and the save procedure is successful
            while not save_atomic_entry(entry,
                                        slot_id,
                                        election_id,
                                        max_nr):

                (slot_id, max_nr) = select_random_entry(UniqueVote,
                                                        election_id)

            # Return True if everything goes according to plan
            return Response(True)

        except Exception:
            # TODO: Log
            return Response(False)


class InitializeElectionView(CreateAPIView):
    """
    Standard API view
    """
    permission_classes = []
    serializer_class = serializers.InitializeElectionSerializer

    def get_serializer_class(self):
        """
        Define the Serializer class to be interpreted
        by Django Swagger
        """
        return self.serializer_class

    def post(self, request, *args, **kwargs):
        """
        POST method docs
        """
        return_data = {"error": "0", "message": "ok"}
        try:
            election_id = self.kwargs.get('election_id') or None
            if election_id is None:
                raise ValidationError("No election ID found")
            entry = ElectionVote.objects.filter(election_id=election_id)[:1]

            if entry.exists():
                raise ValidationError("Election already initialized")

            serializer = serializers.InitializeElectionSerializer(
                                                    data=request.data)
            if(serializer.is_valid(raise_exception=True)):
                sanitized_data = serializer.validated_data
                cardinality = sanitized_data['voters_nr']
                flag = initialize_election(election_id, cardinality)
                if not flag:
                    raise ValidationError("Unexpected error")

        except Exception as e:
            return_data["error"] = "1"
            return_data["message"] = str(e)

        return Response(return_data)


class CloseElectionView(CreateAPIView):
    """
    Standard API view
    """
    permission_classes = []
    serializer_class = serializers.CloseElectionSerializer

    def get_serializer_class(self):
        """
        Define the Serializer class to be interpreted
        by Django Swagger
        """
        return self.serializer_class

    def post(self, request, *args, **kwargs):
        """
        POST method docs
        """
        election_id = self.kwargs.get('election_id') or None
        if election_id is None:
            raise ValidationError("No election ID found")
        entry = ElectionVote.objects.filter(election_id=election_id)[:1]

        if not entry.exists():
            raise ValidationError("Election not initialized")
        serializer = serializers.CloseElectionSerializer(data=request.data)
        if(serializer.is_valid(raise_exception=True)):
            flag = close_election(election_id)
            return Response(flag)


class ListReferencesView(ListAPIView):
    permission_classes = []
    serializer_class = serializers.UniqueVoteSerializer
    filter_backends = (DjangoFilterBackend,)
    filter_class = UniqueVoteFilter
    filter_fields = ('unique_vote_bulletin',)
    pagination_class = StandardPagination

    def get_serializer_class(self):
        """
        Define the Serializer class to be interpreted
        by Django Swagger
        """
        return self.serializer_class

    def get_queryset(self):
        election_id = self.kwargs.get('election_id') or None
        queryset = UniqueVote.objects.all().filter(empty_entry=False,
                                                   election_id=election_id)

        return queryset


class ListVotesView(ListAPIView):
    permission_classes = []
    serializer_class = serializers.ElectionVoteSerializer
    pagination_class = StandardPagination

    def get_serializer_class(self):
        """
        Define the Serializer class to be interpreted
        by Django Swagger
        """
        return self.serializer_class

    def get_queryset(self):
        election_id = self.kwargs.get('election_id') or None
        queryset = ElectionVote.objects.all().filter(empty_entry=False,
                                                     election_id=election_id)

        return queryset


class CountVotesView(GenericAPIView):
    permission_classes = []
    pagination_class = []
    serializer_class = serializers.ElectionVoteSerializer

    def get_serializer_class(self):
        """
        Define the Serializer class to be interpreted
        by Django Swagger
        """
        return self.serializer_class

    def get(self, request, *args, **kwargs):
        return_data = {"count": 0}
        try:
            election_id = self.kwargs.get('election_id') or None
            if election_id is None:
                raise ValidationError("No election ID found")

            serializer = serializers.ElectionVoteSerializer(
                                                    data=request.data)
            if(serializer.is_valid(raise_exception=True)):
                count = ElectionVote.objects.all().filter(
                    empty_entry=False,
                    election_id=election_id
                ).count()
                return_data["count"] = count

        except Exception as e:
            # TODO: Log
            raise e

        return Response(return_data)


class UpdateVotesView(CreateAPIView):
    permission_classes = []
    serializer_class = serializers.UpdateVotesSerializer

    def get_serializer_class(self):
        """
        Define the Serializer class to be interpreted
        by Django Swagger
        """
        return self.serializer_class

    def post(self, request, *args, **kwargs):
        """
        POST method docs
        """
        return_data = {"invalid": 0, "success": [], "failed": []}
        try:
            election_id = self.kwargs.get('election_id') or None
            if election_id is None:
                raise ValidationError("No election ID found")
            entry = ElectionVote.objects.filter(election_id=election_id)[:1]

            if not entry.exists():
                raise ValidationError("Election not initialized")

            serializer = serializers.UpdateVotesSerializer(
                                                    data=request.data)
            if(serializer.is_valid(raise_exception=True)):
                sanitized_data = serializer.validated_data
                votes = sanitized_data['vote_list']
                deciphered_key = 'deciphered_vote'
                for vote in votes:
                    # If there is no deciphered vote mark as invalid
                    deciphered_vote = vote.get(deciphered_key, None)
                    if not deciphered_vote:
                        return_data['invalid'] = return_data['invalid'] + 1
                        continue
                    # If there no ID to update the model, mark as failed
                    vote_entry_id = vote.get('id', None)
                    if not vote_entry_id:
                        return_data["failed"].append(deciphered_vote)
                        continue

                    try:
                        # Get the existing vote from the database
                        vote_obj = ElectionVote.objects.get(pk=vote_entry_id)

                        # Election ID sanity check
                        if int(vote_obj.election_id) != int(election_id):
                            raise ElectionVote.DoesNotExist

                        # Populate the object with the given fields to update
                        for (key, value) in vote.items():
                            setattr(vote_obj, key, value)

                        # Deploy the changes
                        vote_obj.save()

                        # Set as success in the response data
                        return_data["success"].append(vote)

                    except ElectionVote.DoesNotExist:
                        return_data["failed"].append(deciphered_vote)
                        continue

        except Exception as e:
            # TODO: Log
            # print e
            raise e

        return Response(return_data)
